const { debug } = require("./../const");
const Metadata = require("./../metadata");
const Dao = require("./dao/base");
const Datasource = require("./datasource/base");
const DatasourceMap = require("./datasource/manager");
const BoostContext = require("./context");
const Service = require("./service");

const BoostProvider = {
    getService(targetClass) {
        let fields = Metadata.getPropertyMetadata(targetClass.prototype);
        let methods = Metadata.getMethodMetadata(targetClass.prototype);
        let targetInstance = new targetClass();
        return new Proxy(targetInstance, {
            get(target, name) {
                if (typeof target[name] === 'function') {
                    let action = methods[name] || {};
                    let { transaction = false } = action;
                    return (...args) => {
                        let promise = Promise.resolve(), map = new Map();
                        Reflect.ownKeys(fields).filter(a => !!fields[a].dao && fields[a].dao.prototype instanceof Dao).forEach(keyName => {
                            let daoClass = fields[keyName].dao;
                            let { datasource } = Metadata.getClassMetadata(daoClass.prototype) || {};
                            if (datasource && datasource.prototype instanceof Datasource) {
                                let { database } = Metadata.getClassMetadata(datasource.prototype) || {};
                                if (database) {
                                    promise = promise.then(() => {
                                        let _datasource = DatasourceMap.get(database);
                                        return _datasource.getConnection().then(connection => {
                                            target[keyName] = new daoClass(connection);
                                            map.set(_datasource, connection);
                                        });
                                    });
                                }
                            }
                        });
                        return promise.then(() => {
                            let ps = Promise.resolve();
                            if (transaction) {
                                ps = ps.then(() => {
                                    return [...map.values()].reduce((a, connection) => {
                                        return a.then(() => connection.beginTransaction());
                                    }, Promise.resolve()).then(() => {
                                        debug('start transaction');
                                    });
                                });
                            }
                            return ps.then(() => {
                                return new Promise((resolve, reject) => {
                                    let result = true, errorInfo = null, resultInfo = null;
                                    Promise.resolve().then(() => target.beforeExcute(action))
                                        .then(() => target[name].apply(target, args))
                                        .then(a => target.afterExcute({ result: a, action }))
                                        .then(a => {
                                            result = true;
                                            resultInfo = a;
                                            if (transaction) {
                                                return [...map.values()].reduce((a, connection) => {
                                                    return a.then(() => connection.commit());
                                                }, Promise.resolve()).then(() => {
                                                    debug('commit transaction');
                                                });
                                            }
                                        }).catch(err => {
                                            debug(err);
                                            result = false;
                                            errorInfo = err;
                                            if (transaction) {
                                                return [...map.values()].reduce((a, connection) => {
                                                    return a.then(() => connection.rollback());
                                                }, Promise.resolve()).then(() => {
                                                    debug('rollback transaction');
                                                });
                                            }
                                        }).finally(a => {
                                            map.forEach((connection, datasource) => {
                                                datasource.releaseConnection(connection);
                                            });
                                            if (result) {
                                                resolve(resultInfo);
                                            } else {
                                                reject(errorInfo);
                                            }
                                        });
                                });
                            });
                        });
                    }
                } else {
                    return target[name];
                }
            }
        });
    },
    getEntity(targetClass) {
        let targetObject = new targetClass();
        let fields = Metadata.getPropertyMetadata(targetClass.prototype) || {};
        let _service = [], _channel = [], _remoteService = [], _remoteVerify = [];
        Reflect.ownKeys(fields).forEach(propName => {
            if (!!fields[propName].service && fields[propName].service.prototype instanceof Service) {
                _service.push(propName);
            }
            if (!!fields[propName].channel) {
                _channel.push(propName);
            }
            if (!!fields[propName].remoteService) {
                _remoteService.push(propName);
            }
            if (!!fields[propName].remoteVerify) {
                _remoteVerify.push(propName);
            }
        });
        _service.forEach(propName => {
            targetObject[propName] = this.getService(fields[propName].service);
        });
        _channel.forEach(propName => {
            targetObject[propName] = BoostContext.info.channel;
        });
        _remoteService.forEach(propName => {
            targetObject[propName] = BoostContext.info.service;
        });
        _remoteVerify.forEach(propName => {
            targetObject[propName] = BoostContext.info.serviceVerifier;
        });
        return targetObject;
    }
}

module.exports = BoostProvider;